// 3PConv:
// Convert SPIN-binaries to / from files for communication with 3PClient
// compiled with an elder Open-Watcom (
#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "time.h"
#include "sys/types.h"
#include "sys/timeb.h"
#include "assert.h"


#define programString "3PConv V0.1"

typedef enum {undefined, toPropeller, fromPropeller} eDirection_t;
eDirection_t eDirection;

FILE* inFile = NULL;
FILE* outFile = NULL;
char* pInFileName;  // need them for error-messages
char* pOutFileName; 
// --------------------------------------------
int DeNibble(char ch) {
  if ((ch >= '0') & (ch <= '9'))
    return ch - '0';
    
  if ((ch >= 'A') & (ch <= 'F'))
    return ch - 'A' + 10;
    
  if ((ch >= 'a') & (ch <= 'f'))
    return ch - 'a' + 10;
    
  return 0x100; // force an overflow
}
// -------------------------------------------
int ProcessToProp(void) {
// expected input:
// a line starting with '-r:<number1>:<number2>
// or
// a line starting with '-w:<number1>:<number2>
// <number1> is ignored, <number2> is the length>
//
// any line with comments ('//') will be ignored
// following a bunch of lines with (optional <address>':', will be ignored) (up to) 32 hex bytes, a (or more) space and a 1-byte hex checksum
//
// output is simply a binary stream

  #define inBufferSize 200
  char inBuffer[inBufferSize];
  char* pChar;
  long byteCount, length;
  int byte, check;
  
  // get the header ...
  for (;;) {
    if (fgets(inBuffer, inBufferSize, inFile) == NULL) { // premature EOF? Can't read ...
      printf("*** unexpected end of file\n");
      return 0;
    }
    if (strncmp("//", inBuffer, 2) != 0)
      break;
  }

  // decode the header ... inBuffer should contain it

  if ((strncmp("-r:", inBuffer, 3) == 0) | (strncmp("-w:", inBuffer, 3) == 0)) {
    // skip first parameter (address)
    strtol(&inBuffer[3], &pChar, 0);
    length = strtol(++pChar, NULL, 0);
  }
  else {
    printf("*** can't find 3PClient-header in '%s'.\n", pInFileName);
    return 0;
  }
  

  // process lines ...
  byteCount = 0;
  for (;;) {
    // get next line ...
    for (;;) {
      if (fgets(inBuffer, inBufferSize, inFile) == NULL)
        break;
      if (strncmp("//", inBuffer, 2) != 0)
        break;
    };

    // BUGBUG EndOfFile?
     
    // skip address-field
    pChar = strchr(inBuffer, ':');
    if (pChar != NULL)
      pChar++;
    else
      pChar = &inBuffer[0];

    check = 0;
    for (;;) {// one complete line
      byte = (DeNibble(*(pChar++)) << 4) + DeNibble(*(pChar++));
      if (byte > 0xFF) { // got an error
        printf("*** invalid value (non-HEX) near:'%s'\n", inBuffer);
        return 0;
      }

      fputc(byte, outFile);  // BUGBUG: Error-checking
      byteCount++;

//    if (byteCount % 1024 == 0)
//      printf(".");
        
      check += byte;
      if ((byteCount % 16 == 0) | (byteCount > length - 1)) { // end of line, or end of transmition.
        unsigned int lCheck;
        
        // read the checksum, skip any leading blanks in front
        while (*pChar == ' ') 
          pChar++;
          
        lCheck = (DeNibble(*(pChar++)) << 4) + DeNibble(*pChar); // no need for second ++, pChar isn't used anymore

        if (lCheck != check % 0xFF) {
          printf("*** checksum-error 0x%2X : 0x%2X in '%s'\n", check, lCheck, inBuffer);
          return 0;
        }
        
        break;
      }  // one complete line
      
    }

    if (byteCount > length - 1)
      break;    
    
  }

  // BUGBUG: see if we got all the bytes expected
  return 1;
}
// --------------------------------------------
void ProcessFromProp (void) {
// output in 3PClient-format.
// nothing tragic ...
  int length;
  int i;
  int check, byte;
  time_t now;
  
  // find length (= size of file), brute force
  length = 0;
  for (;;) {
    if (fgetc(inFile) == EOF)
      break;
    length++;
  }

  fseek(inFile, 0, SEEK_SET);

  // write the header ...
  time(&now);
  fprintf(outFile, "-w:0x00000:0x%X\x0A", length);
  fprintf(outFile, "// generated by '" programString "' from:'%s' on: %.24s\x0A", pInFileName, asctime(gmtime(&now)));
  
  check = 0;
  for (i = 0; i < length; i++) {
    byte = fgetc(inFile);
    check+= byte;
    fprintf(outFile, "%02X", byte);

    if ((i % 16 == 15) | (i >= length - 1)) {
      fprintf(outFile, " %02X\x0A", check % 0xFF);
      check = 0;
    }
  }
}

// --------------------------------------------
void PrintUsage(void) {
  printf("3PConv: convert Propeller-binaries (*.binary) to/from hex-files for usage with 3PClient\n\n");
  printf("usage:\n");
  printf("3PConv [option] <InputFile> <OutputFile>\n");
  printf("[option]: (mandatory)\n");
  printf("'-p': convert Propeller-binary to 3PClient-Hex-format)\n");
  printf("'-h': convert 3PClient-Hex-format to Propeller-format\n\n");
}
// --------------------------------------------
int OpenInOut(char* inFileName, char* outFileName, eDirection_t direction) {

  if (direction == toPropeller) {
    inFile = fopen(inFileName, "r");
    outFile = fopen(outFileName, "wb");
  }
  else {
    inFile = fopen(inFileName, "rb");
    outFile = fopen(outFileName, "wb");  // needs to be binary to avoid translation of 0x0A into 0x0D0A
  }
  return ((inFile != NULL) & (outFile != NULL));
}
// --------------------------------------------
void CloseInOut(void) {
  if (inFile)
    fclose(inFile);
  if (outFile)
    fclose(outFile);
}
// --------------------------------------------
int main(int argc, char* argv[]) {

  if (argc != 4) {
    PrintUsage();
    return 1;
  }

  eDirection = undefined;
  
  if (strcmp("-p", argv[1]) == 0)
    eDirection = fromPropeller;
  if (strcmp("-h", argv[1]) == 0)
    eDirection = toPropeller;

  if (eDirection == undefined) {
    PrintUsage();
    return 1;
  }

  if (!OpenInOut(argv[2], argv[3], eDirection)) {
    if (inFile == 0)
      printf("unable to open input '%s'\n", argv[2]);
    if (outFile == 0)
      printf("unable to open output '%s'\n", argv[3]);
      
    CloseInOut();

    return 1;
  }

  pInFileName = argv[2];
  pOutFileName = argv[3];


  if (eDirection == toPropeller)
    ProcessToProp();
  else
    ProcessFromProp();

  CloseInOut();
  return 0;
}


